/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 *
 * Original Author: Marc BOULLET
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 
//******** define a js object to implement nsITreeView
function gestCustomizeTreeView(columnids)
{
  // columnids is an array of strings indicating the names of the columns, in order
  this.columnids = columnids;
  this.colcount = columnids.length

  this.copycol = -1;
  this.rows = 0;
  this.tree = null;
  this.data = new Array;
  this.selection = null;
  this.sortcol = null;
  this.sortdir = 0;
  this.stdRowCount = -1;
  this.hoveredRow = -1;
  this.topOrBottom = false;
  this.feedbackBottom = kAtomService.getAtom("aioHovered");
  this.feedbackTop = kAtomService.getAtom("aioHoverCol");
  this.gestEnabled = kAtomService.getAtom("aioGestEnabled");
  this.gestDisabled = kAtomService.getAtom("aioGestDisabled");
  this.before = false;
}

gestCustomizeTreeView.prototype = {
  set rowCount(c) { throw "rowCount is a readonly property"; },
  get rowCount() { return this.rows; },

  setTree: function(tree) {
    this.tree = tree;
  },

  getCellText: function(row, column) {
    return this.data[row][column.index] || "";
  },

  setCellText: function(row, column, value) {
    this.data[row][column.index] = value;
  },

  addRow: function(row) {
    this.rows = this.data.push(row);
    this.rowCountChanged(this.rows - 1, 1);
  },

  removeLastTwoRows: function() {
    this.data.pop();
    this.rowCountChanged(--this.rows, -1);
    this.data.pop();
    this.rowCountChanged(--this.rows, -1);
  },

  rowCountChanged: function(index, count) {
    this.tree.rowCountChanged(index, count);
  },

  invalidate: function() {
    this.tree.invalidate();
  },

  getRowProperties: function(row, prop) {
    if (row == this.hoveredRow)
       if (this.topOrBottom) prop.AppendElement(this.feedbackTop);
       else prop.AppendElement(this.feedbackBottom);
  },

  setCellValue: function(row, column, value) { },	
  handleCopy: function(row) { },
  performActionOnRow: function(action, row) { },

  getCellProperties: function(row, column, prop) {
    if (column.id == "enabledColId")
       if (isEnabledTable[row]) prop.AppendElement(this.gestDisabled);
       else prop.AppendElement(this.gestEnabled);
  },
  getColumnProperties: function(column, prop) { },
  isContainer: function(index) {return false;},
  isContainerOpen: function(index) {return false;},
  isSeparator: function(index) {return index == this.stdRowCount},
  isSorted: function() { },
  canDrop: function(index, orientation) {
     this.before = orientation == -1;
     return false;
  },   
  drop: function(row, orientation) {return false;},
  getParentIndex: function(index) {return -1;},
  hasNextSibling: function(index, after) {return false;},
  getLevel: function(index) {return 0;},
  getImageSrc: function(row, column) { },
  getProgressMode: function(row, column) { },
  getCellValue: function(row, column) { },
  toggleOpenState: function(index) { },
  cycleHeader: function(col) { },
  selectionChanged: function() { },
  cycleCell: function(row, column) { },
  isEditable: function(row, column) {return false;},
  performAction: function(action) { },
  performActionOnCell: function(action, row, column) { }
};

function gestCustomizeTreeViewOld(columnids)
{
  // columnids is an array of strings indicating the names of the columns, in order
  this.columnids = columnids;
  this.colcount = columnids.length

  this.copycol = -1;
  this.rows = 0;
  this.tree = null;
  this.data = new Array;
  this.selection = null;
  this.sortcol = null;
  this.sortdir = 0;
  this.stdRowCount = -1;
  this.hoveredRow = -1;
  this.topOrBottom = false;
  this.feedbackBottom = kAtomService.getAtom("aioHovered");
  this.feedbackTop = kAtomService.getAtom("aioHoverCol");
  this.gestEnabled = kAtomService.getAtom("aioGestEnabled");
  this.gestDisabled = kAtomService.getAtom("aioGestDisabled");
  this.before = false;
}

gestCustomizeTreeViewOld.prototype = {
  set rowCount(c) { throw "rowCount is a readonly property"; },
  get rowCount() { return this.rows; },

  setTree: function(tree) {
    this.tree = tree;
  },

  getCellText: function(row, column) {
    var colidx = 0;
    while(colidx < this.colcount && column != this.columnids[colidx]) ++colidx;
    return this.data[row][colidx] || "";
  },

  setCellText: function(row, column, value) {
    var colidx = 0;
    while(colidx < this.colcount && column != this.columnids[colidx]) ++colidx;
    this.data[row][colidx] = value;
  },

  addRow: function(row) {
    this.rows = this.data.push(row);
    this.rowCountChanged(this.rows - 1, 1);
  },

  removeLastTwoRows: function() {
    this.data.pop();
    this.rowCountChanged(--this.rows, -1);
    this.data.pop();
    this.rowCountChanged(--this.rows, -1);
  },

  rowCountChanged: function(index, count) {
    this.tree.rowCountChanged(index, count);
  },

  invalidate: function() {
    this.tree.invalidate();
  },

  getRowProperties: function(row, prop) {
    if (row == this.hoveredRow)
       if (this.topOrBottom) prop.AppendElement(this.feedbackTop);
       else prop.AppendElement(this.feedbackBottom);
  },

  handleCopy: function(row) { },
  performActionOnRow: function(action, row) { },

  getCellProperties: function(row, column, prop) {
    if (column == "enabledColId")
       if (isEnabledTable[row]) prop.AppendElement(this.gestDisabled);
       else prop.AppendElement(this.gestEnabled);
  },
  getColumnProperties: function(column, elem, prop) { },
  isContainer: function(index) {return false;},
  isContainerOpen: function(index) {return false;},
  isSeparator: function(index) {return index == this.stdRowCount},
  isSorted: function() { },
  canDropOn: function(index) {return false;},
  canDropBeforeAfter: function(index, before) {this.before = before; return false;},
  drop: function(row, orientation) {return false;},
  getParentIndex: function(index) {return -1;},
  hasNextSibling: function(index, after) {return false;},
  getLevel: function(index) {return 0;},
  getImageSrc: function(row, column) { },
  getProgressMode: function(row, column) { },
  getCellValue: function(row, column) { },
  toggleOpenState: function(index) { },
  cycleHeader: function(col, elem) { },
  selectionChanged: function() { },
  cycleCell: function(row, column) { },
  isEditable: function(row, column) {return false;},
  performAction: function(action) { },
  performActionOnCell: function(action, row, column) { }
};

const kAioMime = "text/allinone-row";
const kSAIID = Components.interfaces.nsISupportsArray;
const kSSIID = Components.interfaces.nsISupportsString;
const kDragContractId = "@mozilla.org/widget/dragservice;1";
const kDSIID = Components.interfaces.nsIDragService;
const kXferableContractID = "@mozilla.org/widget/transferable;1";
const kXferableIID = Components.interfaces.nsITransferable;
const kAtomService = Components.classes["@mozilla.org/atom-service;1"].createInstance(Components.interfaces.nsIAtomService);
const kNumberOfWheelGestures = 2;

var gestView, treeBoxView;
var gestureTree, treeBox, editArea, swapArea, funcArea, inputBox, rowCount, rockerCount, totalCount;
var functionCol, gestureCol, enabledCol;
var undoFunc = [], swapFunc = [], undoVal = [], swapVal = [], undoId = [];
var gUndoId = 0;
var swapping = false, editing = true, funcEditing = false, hasBug = false;
var rockerRow = -1, funcRow;
var bundle;
var localizedGest = [];
var abbrLocalizedGest = [];
var selTable = [], selHasRocker = false;
var dragService = Components.classes[kDragContractId].getService(kDSIID);
var abbrTable = [], isEnabledTable = [], funcNbTable = [];
var rockFuncTable = [];

var gestActionTable = [
      "g.browserBack", //0
      "g.browserForward", //1
      "g.browserReload", //2
      "g.browserReloadSkipCache", //3
      "g.browserStop", //4
      "g.browserHome", //5
      "g.openNewWindow", //6
      "g.duplicateWindow", //7
      "g.upDir", //8
      "g.browserOpenTabInFg", //9
      "g.duplicateTab", //10
      "g.nextTab", //11
      "g.previousTab", //12
      "g.closeOther", //13
      "g.restMaxWin", //14
      "g.minWin", //15
      "g.fullScreen", //16
      "g.openSelection", //17
      "g.closeDoc", // 18
      "g.viewPageSource", //19
      "g.viewFrameSource", //20
      "g.viewSiteCookies", //21
      "g.pageInfo", //22
      "g.jsConsole", //23
      "g.about", //24
      "g.addBookmark", //25
      "g.doubleStackWin", //26
      "g.doubleImageSize", //27
      "g.halveImageSize", //28
      "g.hideObject", //29
      "g.zoomIn", //30
      "g.zoomOut", //31
      "g.resetZoom", //32
      "g.w3cValidate", //33
      "g.linksInWindows", //34
      "g.linksInTabs", //35
      "g.metaInfo", //36
      "g.scrollDown", //37
      "g.scrollUp", //38
      "g.scrollToTop", //39
      "g.scrollToBottom", //40
      "g.resetImage", //41
      "g.bookmarkSidebar", //42
      "g.hideFlash", //43
      "g.URLToClipboard", //44
      "g.firstPage", //45
      "g.showGestures", //46
      "g.closeTab", //47
      "g.incURL", //48
      "g.decURL", //49
      "g.clearDigitFlipper", //50
      "g.linksInFiles", //51
      "g.undoCloseTab", //52
      "g.printPreview", //53
      "g.browserOpenTabInBg", //54
      "g.deleteSiteCookies", //55
      "g.undoHideObject", //56
      "g.openFav1", //57
      "g.openFav2", //58
      "g.openBlankTab", //59
      "g.closeWindow", //60
      "g.openWindowInBg", //61
      "g.frameInfo", //62
      "g.aioOptions", //63
      "g.historySidebar", //64
      "g.bookmarkMgr", //65
      "g.translate", //66
      "g.downloadMgr", //67
      "g.savePageAs", //68
      "g.prevSelectedTab", //69
      "g.showHideStatusBar", //70
      "g.reloadFrame", //71
      "g.enlargeObject", //72
      "g.reduceObject", //73
      "g.resetSize", //74
      "g.showHideFindBar", //75
      "g.reloadAllTabs" //76
    ];
var rockerGestName = [
      "g.leftRocker",
      "g.rightRocker",
      "g.forwardWheel",
      "g.backwardWheel"
    ];

function setScrollGesturesVisibility(show) {
  if (rockerCount == rockerGestName.length) {
     if (show) return;
     var j = kNumberOfWheelGestures;
     for (var i = totalCount - kNumberOfWheelGestures; i < totalCount; ++i, j++)
        rockFuncTable[j] = isEnabledTable[i] + funcNbTable[i];
     rockerCount -= kNumberOfWheelGestures; totalCount -= kNumberOfWheelGestures;
     gestView.removeLastTwoRows();
  }
  else {
     if (!show) return;
     rockerCount += kNumberOfWheelGestures;
     j = totalCount;  totalCount += kNumberOfWheelGestures;
     for (i = kNumberOfWheelGestures; i < rockerCount; ++i, ++j) {
        if (rockFuncTable[i].charAt(0) == "/") {
           rockFuncTable[i] = rockFuncTable[i].substr(1);
           isEnabledTable[j] = "/";
        }
        else isEnabledTable[j] = "";
        gestView.addRow([gestActionTable[rockFuncTable[i] - 0], "", rockerGestName[i]]);
        abbrTable[j] = ""; funcNbTable[j] = rockFuncTable[i];
     }
  }
  treeBox.invalidate();
  setTimeout(selectRow, 0, 0);
}

function returnCustomizedString(aCase) {
  var gestTable = [];
  if (hasBug) return "";
  if (aCase == 2) {
     for (var i = rowCount + 1; i < totalCount; ++i) gestTable.push(isEnabledTable[i] + funcNbTable[i]);
     if (rockerCount < rockerGestName.length)
        for (i = rockerCount; i < rockerGestName.length; ++i) gestTable.push(rockFuncTable[i]);
     return gestTable.join("|");
  }
  for (i = 0; i < rowCount; ++i)
     if (aCase) gestTable[i] = funcNbTable[i];
     else gestTable[i] = isEnabledTable[i] + abbrTable[i];
  return gestTable.join("|");
}

function populateTree(aGesturesString, aFuncsString, aRockerString) {
  bundle = document.getElementById("allinonegestbundle")
  rowCount = gestActionTable.length;
  rockerCount = rockerGestName.length;
  totalCount = rowCount + rockerCount + 1;
  for (var i = 0; i < rowCount; ++i)
     gestActionTable[i] = bundle.getString(gestActionTable[i]);
  for (i = 0; i < rockerCount; ++i)
     rockerGestName[i] = bundle.getString(rockerGestName[i]);
  clearSelectionTable();
  localizedGest["R"] = bundle.getString("full.right");
  localizedGest["L"] = bundle.getString("full.left");
  localizedGest["U"] = bundle.getString("full.up");
  localizedGest["D"] = bundle.getString("full.down");
  localizedGest["+"] = bundle.getString("full.any");
  abbrLocalizedGest["R"] = bundle.getString("abbreviation.right");
  abbrLocalizedGest["L"] = bundle.getString("abbreviation.left");
  abbrLocalizedGest["U"] = bundle.getString("abbreviation.up");
  abbrLocalizedGest["D"] = bundle.getString("abbreviation.down");
  abbrLocalizedGest["+"] = "+";

  inputBox = document.getElementById("customTextbox");
  editArea = document.getElementById("customEdit");
  swapArea = document.getElementById("customSwap");
  funcArea = document.getElementById("customEditFunc");
  gestureTree = document.getElementById("gesttree");
  treeBox = gestureTree.treeBoxObject;
  if ("columns" in gestureTree) {
     gestView = new gestCustomizeTreeView(["functionId", "enabledColId", "gestTextId"]);
     functionCol = gestureTree.columns["functionId"];
     gestureCol = gestureTree.columns["gestTextId"];
     enabledCol = gestureTree.columns["enabledColId"];
     treeBoxView = gestView;
  }
  else {
     gestView = new gestCustomizeTreeViewOld(["functionId", "enabledColId", "gestTextId"]);
     functionCol = "functionId";
     gestureCol = "gestTextId";
     enabledCol = "enabledColId";
     treeBoxView = treeBox;
  }
  treeBox.view = gestView;
  abbrTable = aGesturesString.split("|");
  funcNbTable = aFuncsString.split("|");
  rockFuncTable = aRockerString.split("|");
  gestView.stdRowCount = rowCount;
  if (abbrTable.length > rowCount) {
     dump("AiOGest: Number of gestures doesn't match!\n");
     hasBug = true;
  }
  else {
     for (i = 0; i < abbrTable.length; ++i) {
        if (abbrTable[i].charAt(0) == "/") {
           abbrTable[i] = abbrTable[i].substr(1);
           isEnabledTable[i] = "/";
        }
        else isEnabledTable[i] = "";
        gestView.addRow([gestActionTable[funcNbTable[i]-0], "", expandedText(abbrTable[i])]);
     }
     for (i = abbrTable.length; i < rowCount; ++i) {
        gestView.addRow([gestActionTable[i], "", expandedText("")]);
        abbrTable[i] = ""; funcNbTable[i] = i + ""; isEnabledTable[i] = "";
     }
     gestView.addRow(["", "", ""]); // separator
     var j = rowCount + 1;
     for (i = 0; i < rockerCount; ++i, ++j) {
        if (rockFuncTable[i].charAt(0) == "/") {
           rockFuncTable[i] = rockFuncTable[i].substr(1);
           isEnabledTable[j] = "/";
        }
        else isEnabledTable[j] = "";
        gestView.addRow([gestActionTable[rockFuncTable[i] - 0], "", rockerGestName[i]]);
        abbrTable[j] = ""; funcNbTable[j] = rockFuncTable[i];
     }
     setTimeout(selectRow, 0, 0); // some sync
  }
}

function buttEnable(idTable, condTable) {
  var el;
  for (var i = 0; i < idTable.length; ++i) {
    el = document.getElementById(idTable[i]);
    if (condTable[i]) el.removeAttribute("disabled");
    else el.setAttribute("disabled", "true");
  }
}

function clearButtonsButUndo() {
  buttEnable(["clearId", "editId", "swapId", "edfuncId", "undoId"], [false, false, false, false, undoFunc.length]);
}

function expandedText(aString) {
  if (aString.length) {
     var result = "", sep = "";
     for (var i = 0; i < aString.length; ++i) {
        result += sep + localizedGest[aString.charAt(i)];
        sep = "-";
     }
     return result;
  }
  return "----------";
}

function clearSelectionTable() {
  for (var i = 0; i < totalCount; ++i) selTable[i] = false;
}

function selectRow(aRow) {
  treeBoxView.selection.select(aRow);
  treeBox.ensureRowIsVisible(aRow);
}

function getSelections() {
  var sels = [];
  selHasRocker = false;
  var count = treeBoxView.selection.getRangeCount();
  var min = { }, max = { };
  for (var i = 0; i < count; i++) {
     treeBoxView.selection.getRangeAt(i, min, max);
     for (var k = min.value; k <= max.value; ++k)
        if (k != -1) {sels.push(k); selTable[k] = true; if (k >= rowCount) selHasRocker = true;}
  }
  return sels;
}

function getRowFromFuncNb(aNb) {
   for (var i = 0; i < rowCount; ++i)
      if (aNb == funcNbTable[i]) return i;
   dump("AiOGest:" + aNb + " not found in getRow...\n");
   hasBug = true;
   return 0;
}

function setFunctionText(tRow, aRow) {
  treeBox.focused = true;
  selectRow(aRow);
  funcNbTable[aRow] = funcNbTable[tRow];
  gestView.setCellText(aRow, functionCol, gestActionTable[funcNbTable[tRow]]);
  treeBox.invalidateRow(aRow);
}  

function setGestureText(aRow, aString) {
  treeBox.focused = true;
  treeBoxView.selection.select(aRow);
  abbrTable[aRow] = aString;
  gestView.setCellText(aRow, gestureCol, expandedText(aString));
  treeBox.invalidateRow(aRow);
}

function mouseUpInTree(e) {
  gestureTree.removeEventListener("mouseup", mouseUpInTree, true);
  setTimeout(function() {gestureTree.setAttribute("seltype", "multiple");}, 500);
  if (rockerRow != -1) {
     setTimeout(setFunctionText, 100, funcRow, rockerRow);
     rockerRow = -1;
  }
}

function mouseDownInTree(e) {
  if (e.originalTarget.localName != "treechildren") return;
  var r = {}, c = {}, type = {};
  treeBox.getCellAt(e.clientX, e.clientY, r, c, type);
  if (c.value == enabledCol && !e.ctrlKey && !e.shiftKey) {
     if (isEnabledTable[r.value]) isEnabledTable[r.value] = "";
     else isEnabledTable[r.value]= "/";
     treeBox.invalidateCell(r.value, c.value);
     return;
  }
  if (swapping || editing || funcEditing) {
     if (swapping) { // on the selected row since selectionInTree has not been called
        swapping = false;
        undoFunc.pop(); undoVal.pop(); undoId.pop();
        swapArea.setAttribute("hidden", "true");
        buttEnable(["clearId", "editId", "swapId", "undoId"], [true, true, true, undoFunc.length]);
     }
     else if (editing) {
        editing = false;
        editArea.setAttribute("hidden", "true");
        buttEnable(["clearId", "editId", "swapId", "undoId"], [true, true, true, undoFunc.length]);
     }
     else { // funcEditing
        funcEditing = false;
        undoFunc.pop(); undoVal.pop(); undoId.pop(); swapFunc.pop();
        funcArea.setAttribute("hidden", "true");
        buttEnable(["edfuncId", "undoId"], [true, undoFunc.length]);
     }
     gestureTree.addEventListener("mouseup", mouseUpInTree, true);
  }
}

function selectionInTree() {
  var sels = getSelections();
  if (sels.length != 1) {
     if (sels.length && !selHasRocker)
        buttEnable(["clearId", "editId", "swapId", "edfuncId", "undoId"], [true, false, false, false, undoFunc.length]);
     else clearButtonsButUndo();
     return;
  }
  var currRow = sels[0];
  if (swapping || editing || funcEditing) {
     if (editing) {
        editing = false;
        editArea.setAttribute("hidden", "true");
     }
     else if (swapping) {
        swapping = false;
        if (currRow >= rowCount) {
            undoFunc.pop(); undoVal.pop(); undoId.pop();
        }
        else {
           var len = swapFunc.length;
           swapFunc.push(funcNbTable[currRow]);
           var s = abbrTable[currRow];
           swapVal.push(s);
           setGestureText(currRow, undoVal[len]);
           setGestureText(getRowFromFuncNb(undoFunc[len]), s);
        }
        swapArea.setAttribute("hidden", "true");
     }
     else { // funcEditing
        funcEditing = false;
        funcArea.setAttribute("hidden", "true");
        rockerRow = undoFunc[undoFunc.length - 1];
        funcRow = currRow;
        if (currRow == rowCount) {undoFunc.pop(); undoVal.pop(); undoId.pop(); swapFunc.pop(); funcRow = rockerRow;}
     }
     gestureTree.addEventListener("mouseup", mouseUpInTree, true);
  }
  if (currRow < rowCount) buttEnable(["clearId", "editId", "swapId", "edfuncId", "undoId"], [true, true, true, false, undoFunc.length]);
  else if (currRow == rowCount) clearButtonsButUndo();
       else buttEnable(["clearId", "editId", "swapId", "edfuncId", "undoId"], [false, false, false, true, undoFunc.length]);
}

function saveForUndo(aRow, saveId) {
  undoFunc.push(aRow > rowCount ? aRow : funcNbTable[aRow]);
  undoVal.push(aRow > rowCount ? funcNbTable[aRow] : abbrTable[aRow]);
  undoId.push(saveId);
  if (!swapping) {
     swapFunc.push("");
     if (editing) buttEnable(["clearId", "swapId", "undoId"], [true, true, true]);
  }
}

function clearRows() {
  var saveId = ++gUndoId;
  var sels = getSelections();
  for (var i = 0; i < sels.length; ++i) {
     saveForUndo(sels[i], saveId);
     setGestureText(sels[i], "");
  }
}

function editCurrentRow() {
  gestureTree.setAttribute("seltype", "single");
  buttEnable(["clearId", "swapId", "undoId"], [false, false, false]);
  inputBox.value = "";
  editArea.removeAttribute("hidden");
  editing = true;
  setTimeout(setTextBox, 0); // if no sync delay, select() doesn't work
}

function setTextBox() {
  var currRow = getSelections()[0];
  var s = abbrTable[currRow];
  var t = "";
  for (var i = 0; i < s.length; ++i)
     t += abbrLocalizedGest[s.charAt(i)];
  inputBox.value = t;
  if (t) inputBox.select();
  inputBox.focus();
}

function reverseLocalizedGest(aChar) {
  for (var i in abbrLocalizedGest)
    if (aChar == abbrLocalizedGest[i]) return i;
  return "";
}

function openPlayWindow() {
  var rv = { };
  rv.trailColor = document.getElementById("trailPickerId").color;
  rv.trailSize = trailSize;
  rv.mousebutton = document.getElementById("mousebuttOptions").value;
  window.openDialog("chrome://allinonegest/content/pref/aioPlayGesture.xul", "",
       "chrome,centerscreen,modal=yes", rv);
  if (rv.gestString != undefined) inputBox.value = rv.gestString;
  inputBox.select(); inputBox.focus();
}

function newGestValue() {
  var i, s1;
  var s = inputBox.value;
  var t = "", prev = "", currChar;
  var isCorrect = true;
  var currRow = getSelections()[0];
  if (s) {
     for (i = 0; i < s.length; ++i) {
       currChar = reverseLocalizedGest(s.charAt(i).toUpperCase());
       if (currChar && currChar != prev && (currChar != "+"  || i == 0)) {
          t += currChar;
          prev = currChar;
       }
       else {
          alert("'" + s.charAt(i) + "' " + bundle.getString("error.invalidChar") + " " + (i+1));
          isCorrect = false;
          break;
       }
     }
     if (isCorrect && t.charAt(0) == "+" && t.length != 3 && t.length != 4) {
        alert(bundle.getString("error.plusChar"));
        isCorrect = false;
     }
     if (isCorrect) { // check if not duplicated or conflicting
        for (i = 0; i < rowCount; ++i) {
            s1 = abbrTable[i];
            if (t == s1  || (s1.charAt(0) == "+" && t.length >= s1.length &&
                      t.substr(-s1.length + 1) == s1.substr(-s1.length + 1))) break;
        }
        if (i != rowCount && i != currRow)
           if (t == s1)
              if (confirm(s.toUpperCase() + " " + bundle.getString("error.conflict") +
                    gestView.getCellText(i, functionCol) + "'\n" + bundle.getString("error.swap"))) {
                 saveForUndo(currRow, ++gUndoId);
                 swapFunc.push(funcNbTable[i]);
                 swapVal.push(t);
                 setGestureText(i, abbrTable[currRow]);
                 setGestureText(currRow, t);
                 gestureTree.setAttribute("seltype", "multiple");
                 editArea.setAttribute("hidden", "true");              
                 return;
              }
              else isCorrect = false;
           else {
              alert (s.toUpperCase() + " " + bundle.getString("error.conflict") +
                       gestView.getCellText(i, functionCol) + "'");
              isCorrect = false;
           }
     }
  }
  if (isCorrect) {
     editArea.setAttribute("hidden", "true");
     saveForUndo(currRow, ++gUndoId);
     setGestureText(currRow, t);
     gestureTree.setAttribute("seltype", "multiple");
  }
  else setTextBox();
}

function swapTwoRows() {
  gestureTree.setAttribute("seltype", "single");
  buttEnable(["clearId", "editId", "swapId", "undoId"], [false, false, false, false]);
  swapArea.removeAttribute("hidden");
  swapping = true;
  saveForUndo(getSelections()[0], ++gUndoId);
}

function editFunction() {
  gestureTree.setAttribute("seltype", "single");
  buttEnable(["edfuncId", "undoId"], [false, false]);
  funcArea.removeAttribute("hidden");
  funcEditing = true;
  saveForUndo(getSelections()[0], ++gUndoId);
}

function performUndo() {
  var row, saveId, swFunc;
  do {
    saveId = undoId.pop();
    swFunc = swapFunc.pop();
    if (swFunc) setGestureText(getRowFromFuncNb(swFunc), swapVal.pop());
    row = undoFunc.pop();
    if (row < rowCount) {
       row = getRowFromFuncNb(row);
       setGestureText(row, undoVal.pop());
    }
    else setFunctionText(getRowFromFuncNb(undoVal.pop()), row);
  } while (undoId.length && undoId[undoId.length - 1] == saveId)
  treeBox.ensureRowIsVisible(row);
  buttEnable(["undoId"], [undoFunc.length]);
}

function eraseDropFeedback() {
  var rowToInvalidate = gestView.hoveredRow;
  if (rowToInvalidate < 0) return;
  gestView.hoveredRow = -1; gestView.topOrBottom = false;
  treeBox.invalidateRow(rowToInvalidate);
}

function onBeginRowDrag(e) {
  clearSelectionTable();
  var sels = getSelections();
  if (e.originalTarget.localName != "treechildren" || !dragService || selHasRocker) return;
  var transArray = Components.classes["@mozilla.org/supports-array;1"].createInstance(kSAIID);
  if (!transArray) return;
  var trans = Components.classes[kXferableContractID].createInstance(kXferableIID);
  if (!trans) return;
  gestView.hoveredRow = -1; gestView.topOrBottom = false;
  trans.addDataFlavor(kAioMime);
  var rowData = Components.classes["@mozilla.org/supports-string;1"].createInstance(kSSIID);
  rowData.data = sels.join();
  trans.setTransferData(kAioMime, rowData, rowData.data.length * 2);
  transArray.AppendElement(trans.QueryInterface(Components.interfaces.nsISupports));

  dragService.invokeDragSession(e.target, transArray, null, kDSIID.DRAGDROP_ACTION_COPY + kDSIID.DRAGDROP_ACTION_MOVE);
  e.preventBubble();
  eraseDropFeedback(); // clean-up if something went wrong
}

function onDragOutTree(e) {
  eraseDropFeedback();
}

function onDragOverTree(e) {
  if (!dragService) return;
  var dragSession = dragService.getCurrentSession();
  if (!dragSession || !dragSession.isDataFlavorSupported(kAioMime)) return;
  var dragRow = -1; var top = false;
  if (e.originalTarget.localName == "treechildren") {
     dragRow = treeBox.getRowAt(e.clientX, e.clientY);
     dragSession.canDrop = dragRow >= 0 && dragRow < rowCount && !selTable[dragRow];
     if (dragRow >= rowCount || selTable[dragRow]) dragRow = -1;
     else if (!dragRow) top = gestView.before;
  }
  if (dragRow != gestView.hoveredRow || top != gestView.topOrBottom) {
     var rowToInvalidate = gestView.hoveredRow;     
     gestView.hoveredRow = dragRow; gestView.topOrBottom = top;
     if (rowToInvalidate >= 0) treeBox.invalidateRow(rowToInvalidate);
     if (gestView.hoveredRow >= 0) treeBox.invalidateRow(gestView.hoveredRow);
  }
  e.preventBubble();
}

function onDropOnTree(e) {
  if (!dragService) return;
  var dragSession = dragService.getCurrentSession();
  if (!dragSession) return;
  var trans = Components.classes[kXferableContractID].createInstance(kXferableIID);
  if (!trans) return;
  trans.addDataFlavor(kAioMime);
  dragSession.getData(trans, 0);
  var dataObj = {}, bestFlavor = {}, len = {};
  trans.getAnyTransferData(bestFlavor, dataObj, len );
  if (dataObj) dataObj = dataObj.value.QueryInterface(kSSIID);
  if (!dataObj) return;
  var rowStr = dataObj.data.substring(0, len.value / 2).split(",");
  var dragDropRow = treeBox.getRowAt(e.clientX, e.clientY) - (gestView.topOrBottom - 0);
  //move the rows
  for (var i = 0; i < rowStr.length; ++i) {
    if (rowStr[i] >= dragDropRow) break;
    setTimeout(performRowsMove, 1, dragDropRow, rowStr[i] - i);
  }
  for (var j = i; j < rowStr.length; ++j)
      setTimeout(performRowsMove, 1, dragDropRow++, rowStr[j]);
  e.preventBubble();
}

function aioGetRowValue(aRow) {
  var rowObj = {fTxt: "", gTxt: "", aTxt: "", fNbr: "", eChr: ""};
  rowObj.fTxt = gestView.getCellText(aRow, functionCol);
  rowObj.gTxt = gestView.getCellText(aRow, gestureCol);
  rowObj.aTxt = abbrTable[aRow];
  rowObj.fNbr = funcNbTable[aRow];
  rowObj.eChr = isEnabledTable[aRow];
  return rowObj;
}

function aioSetRowValue(aRow, rowObj) {
  gestView.setCellText(aRow, functionCol, rowObj.fTxt);
  gestView.setCellText(aRow, gestureCol, rowObj.gTxt);
  abbrTable[aRow] = rowObj.aTxt;
  funcNbTable[aRow] = rowObj.fNbr;
  isEnabledTable[aRow] = rowObj.eChr;
}

function performRowsMove(dragEndRow, dragStartRow) {
  var i;
  var initialRowObj = aioGetRowValue(dragStartRow);
  if (dragStartRow <= dragEndRow)
    for (i = dragStartRow; i < dragEndRow; ++i) aioSetRowValue(i, aioGetRowValue(i + 1));
  else {
     ++dragEndRow;
     for (i = dragStartRow; i > dragEndRow; --i) aioSetRowValue(i, aioGetRowValue(i - 1));
  }
  aioSetRowValue(dragEndRow, initialRowObj);
  gestView.hoveredRow = -1;
  treeBox.invalidate();
  setTimeout(selectRow, 0, dragEndRow);
}
