/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mouse Gesture for Mozilla.
 *
 * The Initial Developer of the Original Code is Jens Bannmann.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s) (alphabetical order):
 *  Jens Bannmann <jens.b@web.de>
 *  Mike Kozlowski <mozdev@klio.org>
 *  Modified by M. Boullet for All-in-One gestures
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var aioTrailDot;
var aioTrailCont = null;
var aioTrailCnt;
var aioTrailX, aioTrailY, aioDocX, aioDocY;


function aioStartTrail(e){
  var targetDoc = e.target.ownerDocument;
  if (targetDoc.contentType == "application/vnd.mozilla.xul+xml" ||
     (targetDoc.contentType.indexOf("/xml") != -1 && targetDoc.styleSheets && 
      targetDoc.styleSheets[0].href == "chrome://global/content/xml/XMLPrettyPrint.css")) return;
  if (targetDoc.defaultView.top instanceof Window) targetDoc = targetDoc.defaultView.top.document;
  var insertionNode = (targetDoc.documentElement) ? targetDoc.documentElement : targetDoc;
  var docBox = targetDoc.getBoxObjectFor(insertionNode);
  aioDocX = docBox.screenX - targetDoc.defaultView.pageXOffset;
  aioDocY = docBox.screenY - targetDoc.defaultView.pageYOffset;
  aioTrailCont = targetDoc.createElementNS(xhtmlNS, "aioTrailContainer");
  insertionNode.appendChild(aioTrailCont);

  aioTrailDot = targetDoc.createElementNS(xhtmlNS, "aioTrailDot");
  aioTrailDot.style.width = aioTrailSize + "px";
  aioTrailDot.style.height = aioTrailSize + "px";
  aioTrailDot.style.background = aioTrailColor;
  aioTrailDot.style.border = "0px";
  aioTrailDot.style.position = "absolute";
  aioTrailDot.style.zIndex = "10000";
  aioTrailX = e.screenX;
  aioTrailY = e.screenY;
  aioTrailCnt = 0;
}

function aioDrawTrail(e) {
  function appendDot(x, y) {
    if ((++aioTrailCnt & 1) || aioTrailSize == 1) {
      var dot = aioTrailDot.cloneNode(true);
      dot.style.left = x + "px";
      dot.style.top = y + "px";
      aioTrailCont.appendChild(dot);
    }
  }
  if (!aioTrailCont) return;
  var xMove = e.screenX - aioTrailX;
  var yMove = e.screenY - aioTrailY;
  var xDecrement = xMove < 0 ? 1 : -1;
  var yDecrement = yMove < 0 ? 1 : -1;
  var i, currX = e.screenX - aioDocX, currY = e.screenY - aioDocY;
  if (Math.abs(xMove) >= Math.abs(yMove))
    for (i = xMove; i != 0; i += xDecrement)
      appendDot(currX - i, currY - Math.round(yMove * i / xMove));
  else
    for (i = yMove; i != 0; i += yDecrement)
      appendDot(currX - Math.round(xMove * i / yMove), currY - i);
  aioTrailX = e.screenX; aioTrailY = e.screenY;
}

function aioEraseTrail() {
  if (aioTrailCont && aioTrailCont.parentNode)
    aioTrailCont.parentNode.removeChild(aioTrailCont);
  aioTrailCont = null;
}
